#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This Python script is provided as supporting information to the manuscript entitled 
"Non-idealities in lab-scale kinetic testing: a theoretical study of a modular Temkin reactor"
by Gregor D. Wehinger, Bjarne Kreitz, and C. Franklin Goldsmith
The script can be used to simulate the concentration profiles in the Temkin reactor using a CRN consisting of
a PFR with inert and reactive zones as well as back-mixing.
A routine is implemented that fits the recycle ratio to the CFD simulation results
"""

import csv
import cantera as ct
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import matplotlib.gridspec as gridspec
from matplotlib.ticker import (MultipleLocator, AutoMinorLocator)
import string

#Matplotlib settings

labelsizevar = 14
plt.rcParams['figure.figsize']=(5,8)
plt.rcParams['axes.linewidth'] = 2.0
plt.rcParams['lines.linewidth'] = 3.0
plt.rc('xtick', labelsize=14)
plt.rc('ytick', labelsize=14)
plt.rc('axes', labelsize=labelsizevar+2)
plt.rc('legend', fontsize=labelsizevar)
plt.rcParams['lines.markersize'] = 10
plt.rcParams['xtick.direction']='in'
plt.rcParams['ytick.direction']='in'
plt.rcParams['xtick.major.size']=10
plt.rcParams['xtick.major.width']=2
plt.rcParams['xtick.major.size']=10
plt.rcParams['xtick.minor.width']=1
plt.rcParams['xtick.minor.size']=5
plt.rcParams['ytick.major.width']=2
plt.rcParams['ytick.major.size']=10
plt.rcParams['ytick.minor.width']=1
plt.rcParams['ytick.minor.size']=5
plt.rcParams['legend.edgecolor']='k'
plt.rcParams['axes.unicode_minus']=False
plt.rcParams["legend.framealpha"] = 1
plt.rcParams['xtick.major.pad'] = 8
plt.rcParams['ytick.major.pad'] = 8
plt.rcParams['legend.handletextpad']=0.4
plt.rcParams['legend.columnspacing']=0.5
plt.rcParams['legend.labelspacing']=0.3
plt.rcParams['font.family'] = 'sans-serif'#'Arial'
colormap=plt.cm.Dark2 # tab10 #Set2 Set1 Dark2 tab10
colors = [colormap(i) for i in np.linspace(0, 1, 8)]

## Import the CFD data
#      mole fraction on streamline / %
CFDfile1 = r'T700K100mlmin-MoleFractionOnStreamline.csv' # Data file
dfus = pd.read_csv(CFDfile1) # files as is
dF = dfus.sort_values('Copy of Streamline: Direction [-1,0,0] (m)', 
                      ascending=True) # sorted file
x_dist = dF['Copy of Streamline: Direction [-1,0,0] (m)'].to_numpy()
x_dist = (x_dist-x_dist.min())*1000-2.06507029
x_H2 = dF['Copy of Streamline: Mole Fraction of H2(4)'].to_numpy()
x_CH4 = dF['Copy of Streamline: Mole Fraction of CH4(1)'].to_numpy()
x_CO = dF['Copy of Streamline: Mole Fraction of CO(5)'].to_numpy() 
x_CO2 = dF['Copy of Streamline: Mole Fraction of CO2(2)'].to_numpy()
x_H2O = dF['Copy of Streamline: Mole Fraction of H2O(3)'].to_numpy()

#Cut of the first few points of the arrays since these are not of interest
x_dist=x_dist[11:]
x_H2=x_H2[11:]
x_CH4=x_CH4[11:]
x_CO=x_CO[11:]
x_CO2=x_CO2[11:]
x_H2O=x_H2O[11:]

#Here is the long list of parameters
T=700 #temperature in K
length = 9.961e-3  # Catalyst bed length in m
Vdot = 100 # volume flow in mlN/min
vflow = Vdot *1e-6/60 # volume flow in  m**3/s
pressure=1e5 #pressure
vol = 1.291e-7 # m**3, reactor volumen of the Temkin segment
A_geo = 5.022e-5 # m**2, geometric area of the catalyst
F_catgeo = 92 # catalyst to geometry area ratio
A_cat = F_catgeo * A_geo # m**2, catalytic area catalyst
cat_area_per_vol = A_cat/vol  # Catalyst particle surface area per unit volume

#Create an output file and write everything in it. Only the last run is finally saved
output_filename='PFR_Temkin700K100mlmin.txt'

#######################################################################
# Input Parameters  - Sphere
#######################################################################

def Fit(a):
    global solution   

    #input file containing the surface reaction mechanism
    file="".join(('chem.cti'))
    cti_file = file
    
    gas = ct.Solution(cti_file, 'gas')
    gas.TPX = T, pressure, 'Ar:0.0, CH4(1):0.0, H2O(3):0.0, CO2(2):0.2, H2(4):0.8, CO(5):0.0'

    # import the surface model
    surf = ct.Interface(cti_file,'surface1', [gas])
    surf.TP = T, pressure
    surf.coverages = {'site(7)':1} # Nickel surface
       
    #This initializes the phase for the backmixing stream
    #We use the global variable solution to update this with the exit gas concentration 
    #from the previous run
    mixture = ct.Solution(cti_file, 'gas')
    mixture.TPX = T, pressure, solution
        
    
    outfile = open(output_filename,'w')
    writer = csv.writer(outfile)
    writer.writerow(['dist'] + gas.species_names + surf.species_names)

    # The PFR will be simulated by a chain of 'NReactors' stirred tank reactors.
    NReactors = 101

    #The catalyst start and end point (Parameter from CFD geometry)
    begin_cat = 2.5E-3
    on_catalyst = int(begin_cat/length * NReactors)  
    end_cat = 7.6E-3
    off_catalyst =  int(end_cat/length * NReactors)

    dt = 1.0
        
    rlen = length/(NReactors-1)
    rvol = vol/(NReactors-1)

    # catalyst area in one reactor
    cat_area = cat_area_per_vol * rvol
    mass_flow_rate =  (1-a)*vflow* gas.mean_molecular_weight*pressure/ct.gas_constant/273.15
    
    r = ct.IdealGasReactor(gas, energy='off')
    r.volume = rvol
    upstream = ct.Reservoir(gas, name='upstream')
    downstream = ct.Reservoir(gas, name='downstream')
    
    #The entire active surface area needs is in the active part of the PFR
    #Consequently, we must increase this value by the factor alpha
    alpha = length / (end_cat - begin_cat) 
    rsurf = ct.ReactorSurface(surf, r, A=cat_area*alpha)
    
    #mass flow controller for the mixing chamber, this contains "a" the parameter we fit
    mass_flow_rate_mixer =  a*vflow* mixture.mean_molecular_weight*pressure/ct.gas_constant/273.15
    
    #mass flow controller from the upstream reservoir to the mixing chamber
    mfc = ct.MassFlowController(upstream, r, mdot=mass_flow_rate)

    #This is the valve from the reactor to the exit reservoir
    outlet= ct.PressureController(r, downstream, master=mfc, K=1e-9)
    
    sim = ct.ReactorNet([r])
    sim.max_err_test_fails = 1000
    
    ## set relative and absolute tolerances on the simulation
    sim.rtol = 1.0e-12
    sim.atol = 1.0e-22
    
    #Create arrays for the results
    sim_H2=np.zeros(NReactors)
    sim_CO2=np.zeros(NReactors)
    sim_CH4=np.zeros(NReactors)
    sim_CO=np.zeros(NReactors)
    sim_H2O=np.zeros(NReactors)
    sim_dist=np.zeros(NReactors)
    
    #Compute the mass flow rate of each species mi for the recycle stream
    mi_recycle=mixture.Y*mass_flow_rate_mixer

    #Create a quantity of the mass flow rate that we can update during the iteration
    adjust_mass_flow_rate=mass_flow_rate
    
    surf.set_multiplier(0)
    for n in range(NReactors):
                #Compute the mass flow rate of each species mi from the previous reactor
                #outlet.mdot gives the mass flow rate at the outlet of the reactor
                mi_reactor_exit=r.thermo.Y*outlet.mdot(1)
                #Sum of the recycle and reactor mass flow rate of each species
                mi_mixed=mi_reactor_exit+mi_recycle
                #Compute the new mass fractions
                yi_mixed=mi_mixed/(outlet.mdot(1)+mass_flow_rate_mixer)
                #We only want to have back-mixing in the initial part of the reactor
                if n <on_catalyst:
                    #Add the mass flow rate of the mixer mass flow controller mfc
                    adjust_mass_flow_rate+=mass_flow_rate_mixer
                    # Set the state of the reservoir to match the recycle mixture
                    gas.Y = yi_mixed
                else:
                    #We don't have any back-mixing after the initial part
                    adjust_mass_flow_rate+=0
                    # Set the state of the reservoir to match that of the previous reactor
                    gas.TDY=r.thermo.TDY

                upstream.syncState() 
                #Set the new mass flow to the mass flow controller
                mfc.set_mass_flow_rate(adjust_mass_flow_rate)    
                
                #gas.TDY = r.thermo.TDY
                #upstream.syncState()
                if n==on_catalyst:
                    surf.set_multiplier(1)
                if n==off_catalyst:
                    surf.set_multiplier(0)
                sim.reinitialize()
                sim.advance_to_steady_state(max_steps=1e5)
                dist = n * rlen * 1.0e3   # distance in mm 
                sim_dist[n]=dist
                sim_CH4[n]=gas.X[4]
                sim_CO[n]=gas.X[8]
                sim_H2O[n]=gas.X[6]
                sim_CO2[n]=gas.X[5]
                sim_H2[n]=gas.X[7]
                writer.writerow([dist] + list(gas.X) + list(surf.X)) 

    #Update the solution array with the gas phase concentration at the reactor exit            
    solution=gas.X           

    writer.writerow([dist] + list(gas.X) + list(surf.X))  
    outfile.close()    
            
    #This routine performs the calculation of the root mean square error (Sim-CFD)
    idx=np.zeros((len(x_CH4)))
    CH4_comparison=np.zeros((len(x_CH4)))
    CO_comparison=np.zeros((len(x_CH4)))
    H2O_comparison=np.zeros((len(x_CH4)))
    rmse=np.zeros((len(x_CH4)))
    dist_comparison=np.zeros(len(x_CH4))            
    
    for i in range(len(x_CH4)):
        #We need to find the dist (x-value) in best agreement with the CFD value
        #since the dist (x-value) in the CFD and the simulation don't exactly match, 
        idx[i] = np.argmin(np.abs(np.array(sim_dist)-(x_dist[i])))
        
        #Get the concentrations for the x-value we want to compare with the CFD result
        CH4_comparison[i]=sim_CH4[int(idx[i])]
        CO_comparison[i]=sim_CO[int(idx[i])]
        H2O_comparison[i]=sim_H2O[int(idx[i])]
        dist_comparison[i]=sim_dist[int(idx[i])]
        
        #Compute the RMSE for the all product concentrations
        rmse[i]=0
        rmse[i]+=np.sqrt((x_CH4[i]-CH4_comparison[i])**2)
        rmse[i]+=np.sqrt((x_CO[i]-CO_comparison[i])**2)
        rmse[i]+=np.sqrt((x_H2O[i]-H2O_comparison[i])**2)

    rmse_sum=sum(rmse) 
    recycle_ratio=(on_catalyst-1)*a
    print("Recycle ratio:", recycle_ratio)
    
    return rmse_sum

#Initial guess for the exit concentraion. Is updated between iterations
solution=[0., 0. , 0. ,0., 0.00362943, 0.18375966, 0.02132154, 0.77722669, 0.01406268]
#Initial guess for the backmixing ratio
x0=np.array([0.001])
from scipy.optimize import minimize_scalar
res=minimize_scalar(Fit,x0,bounds=(0.00,0.1), method='bounded')

#Plot the results
data=pd.read_csv(output_filename, sep=",", header=0)
sim_dist = data['dist']
sim_CH4 = data['CH4(1)']
sim_CO = data['CO(5)']
sim_H2O = data['H2O(3)']
sim_CO2 = data['CO2(2)']
sim_H2 = data['H2(4)']
fig = plt.figure(constrained_layout=True, dpi=330)
spec = gridspec.GridSpec(ncols=1, nrows=2, figure=fig)
ax1=fig.add_subplot(spec[1, 0])
ax0=fig.add_subplot(spec[0, 0], sharex=ax1)
plt.setp(ax0.get_xticklabels(), visible=False)
#### REACTANTS
ax0.text(-0.21, 1.0, 'A', transform=ax0.transAxes, 
        size=16, weight='bold')
ax0.set_xlim([0, 10])
ax0.set_ylim([0, 100])
ax0.xaxis.set_minor_locator(MultipleLocator(.5))
ax0.yaxis.set_minor_locator(MultipleLocator(10.))
ax0.set_ylabel('$\mathrm{mole\ fraction\ /\ \%}$')
ax0.plot([],[], color='k', label='CFD model', linestyle='dashed')
ax0.plot([],[], color='k', label='CRN model')
ax0.plot(sim_dist, sim_H2*100,  color=colors[1],                label='$\mathrm{H_2}$')
ax0.plot(x_dist, x_H2*100, color=colors[1], linestyle = 'dashed')
ax0.plot(sim_dist, sim_CO2*100,  color=colors[0],           label='$\mathrm{CO_2}$')
ax0.plot(x_dist, x_CO2*100, color=colors[0], linestyle = 'dashed')
ax0.legend(loc='best')   
ax0.text(0.03, 0.94, r'Sphere', verticalalignment='baseline', horizontalalignment='left',
        transform=ax0.transAxes, color='k')
ax0.text(0.03, 0.64, r'$T=$%5.1f K' %T, verticalalignment='baseline', horizontalalignment='left',
        transform=ax0.transAxes, color='red')
ax0.text(0.03, 0.57, r'$\dot V=$%5.1f ml$_{N}$/min' %Vdot, verticalalignment='baseline', horizontalalignment='left',
        transform=ax0.transAxes, color='red')
ax0.text(0.03, 0.50, r'$F_\mathrm{cat/geo}=$%5.1f' %F_catgeo, verticalalignment='baseline', horizontalalignment='left',
        transform=ax0.transAxes, color='red')
#### PRODUCTS
ax1.text(-0.21, 1.0, 'B', transform=ax1.transAxes, 
        size=16, weight='bold')
ax1.set_xlim([0, 10])
ax1.set_ylim([0, 6.])
ax1.xaxis.set_minor_locator(MultipleLocator(.5))
ax1.yaxis.set_minor_locator(MultipleLocator(.50))
ax1.set_ylabel('$\mathrm{mole\ fraction\ /\ \%}$')   
ax1.set_xlabel('$\mathrm{length\ /\ mm}$')
ax1.plot(sim_dist, sim_CH4*100,  color=colors[2],                label='$\mathrm{CH_4}$')
ax1.plot(x_dist, x_CH4*100, color=colors[2], linestyle = 'dashed')
ax1.plot(sim_dist, sim_CO*100,  color=colors[4],           label='$\mathrm{CO}$')
ax1.plot(x_dist, x_CO*100, color=colors[4], linestyle = 'dashed')
ax1.plot(sim_dist, sim_H2O*100,  color=colors[3],  label='$\mathrm{H_2O}$')
ax1.plot(x_dist, x_H2O*100, color=colors[3], linestyle = 'dashed')
ax1.legend()   
plt.savefig('profiles700K100mlminRecycleT.pdf', dpi=300, bbox_inches='tight', transparent=False)